*******************************************************************************;
** SAS macro DropMV   vs. 0.2.2 , 23 Mar 2005, by Jim Groeneveld, Netherlands *;
*******************************************************************************;
**                                                                            *;
*; %MACRO   DROPMV     /* Drop Missing Variables with missing values only     */
/* -------------------------------------------------------------------------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
(  SourceDs=           /* Source data set specification                     R */
,  TargetDs=           /* Target data set specification (may be SourceDs)   R */
,  OverWrit=N          /* specify Y to overwrite an existing TargetDs       O */
,  ExclList=           /* list with variable names to exclude from removal  O */
,  ChkExist=Y          /* check for existence of ExclList variables: Y/N    O */
,  Version = 0.2.2     /* forced version specification for version control  O */
,  Debug   = 0         /* Debug level; for testing, debugging purposes only O */
,  CalledBy=           /* Macro name of macro or program for err0r report   O */
,  ForMyUse=           /* argument for the author's use only, do not use it ! */
)/ DES     =             'Drop redundant missing values variables' /* STORE */ ;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* File       : DROPMV.SAS / DROPMV.022                                       *;
%*                                                                            *;
%* Purpose    : Removal of variables with missing values only                 *;
%*                                                                            *;
%* Programmer : Jim Groeneveld                                                *;
%* Date       : 23 Mar 2005                                                   *;
%* Version    : 0.2.2                                                         *;
%*                                                                            *;
%* SAS version: 6.12 under Windows 95 / 8.2 under Windows W2K, WXP            *;
%*                                                                            *;
%* Validation : - (not yet validated, check results, use at your own risk)    *;
%* Disclaimer : The author is not liable in any way for any negative, adverse *;
%*              consequences of the use or misuse of this program / macro.    *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Notes:                                                                     *;
%* ======                                                                     *;
%* DropMV, as from vs. 0.2.2, also drops numerical variables with all missing *;
%* values equal to any of ._, ., and .a to .z from a dataset, only if all of  *;
%* its values have exactly the same missing value. If a numerical variable    *;
%* has different missing values, it will not be dropped (future option).      *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* History of DROPMV:                                                         *;
%* ==================                                                         *;
%* version  date         description of changes, additions, improvements      *;
%* -------  -----------  ---------------------------------------------------- *;
%* 0.0      19 Jun 2001  1st working release                                  *;
%* 0.1      28 Jun 2001  added ChkExist argument                              *;
%* 0.11     13 Apr 2004  adapted PROC CONTENTS Name into UPCASE(Name) because *;
%*                          of SAS 8.2 backwards incompatibility with 6.12    *;
%* 0.2       8 Jun 2004  removed _LAST_ default value for SourceDs, because   *;
%*                         of inapppropriate processing,                      *;
%* 0.21      9 Jun 2004  corrected macro description                          *;
%* 0.2.2    23 Mar 2005  changed single numerical missing . in to all 28 ones,*;
%*                       added CalledBy argument, Version and Debug control,  *;
%*                       added _LAST_ keyword with SourceDs and TargetDS again*;
%*                                                                            *;
%* Future features:                                                           *;
%* ================                                                           *;
%* - store state of options MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN    *;
%*      to reset them at their original state at the end if changed to debug. *;
%* - either use %SAS_name or code as in %MR2RM to check for valid SAS names.  *;
%* - optionally drop numerical variables with different missing values.       *;
%*                                                                            *;
%******************************************************************************;

%* OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN;
%* remove comment chars for testing ;

  %IF ( %UPCASE(&ForMyUse) = COMPILE ) %THEN
  %DO;
    %GOTO Finish;
  %END;

%IF (&Debug EQ 1 OR &Debug GT 11) %THEN %DO;
  OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN;
%END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* LOCAL macro variables declarations;
%*___________________________________;

 %LOCAL    /* Define internal macro variables explicitely local */
  MacName  /* This macros external name: DROPMV */
  MacroVs  /* Version number major.minor.patch of macro: 0.2.2 */
  MacDate  /* Date of this version of macro: 23 Mar 2005 */
  I        /* element index */
  InclList /* list of variables to include in potential removal */
  Variable /* element of InclList */
  Contents /* stand-in name for Contents dataset ___Cont_ */
  Scratch  /* stand-in name for Scratch datasets ___Scrat */
  DropList /* list of empty variables to remove actually */
  ErrCount /* number of erroneously specified arguments */
  Err      /* text 'E R R O R' as consecutive characters */
  Warn     /* text 'W A R N I N G' as consecutive characters */
  ;

%* Stand-in variable and dataset names;
  %LET MacName  = DROPMV;
  %LET MacroVs  = 0.2.2  ;
  %LET MacDate  = 23 Mar 2005;
  %LET Contents = ___Cont_;
  %LET Scratch  = ___Scrat;
  %LET Err      = %SYSFUNC(COMPRESS(E R R O R, %STR( )));
  %LET Warn     = %SYSFUNC(COMPRESS(W A R N I N G, %STR( )));

%* Macro header in LOG file, dependent on value of CalledBy;
 %IF ( %UPCASE(&CalledBy) EQ %UPCASE(&MacName) OR &CalledBy EQ ) %THEN
 %DO;
  %PUT;
  %PUT |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|;
  %PUT |   Macro &MacName, vs &MacroVs, by Jim Groeneveld, &MacDate   |;
  %PUT |      Drop Missing Variables with missing values only       |;
  %PUT |____________________________________________________________|;
  %PUT;

%* Feedback of specified arguments;
  %PUT &MacName: SourceDs = &SourceDs;
  %PUT &MacName: TargetDs = &TargetDs;
  %PUT &MacName: OverWrit = &OverWrit;
  %PUT &MacName: ExclList = &ExclList;
  %PUT &MacName: ChkExist = &ChkExist;
  %PUT &MacName: Version  = &Version ;
  %PUT &MacName: Debug    = &Debug   ;
  %PUT &MacName: CalledBy = &CalledBy;
  %*PUT &MacName: ForMyUse = &ForMyUse;
 %END;

  %IF (&Version NE AND &Version NE &MacroVs) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Specified version &Version does not match &MacroVs,;
    %PUT .                       macro &MacName will abort;
    %GOTO Finish;
  %END;

%* Process keyword _LAST_ with both SourceDs and TargetDs;
  %IF (%UPCASE(&SourceDs) EQ _LAST_) %THEN %LET SourceDs = &SYSLAST;
  %IF (%UPCASE(&TargetDs) EQ _LAST_) %THEN %LET TargetDs = &SYSLAST;

%* Convert lower case letters to upper case;
  %IF ( &OverWrit NE ) %THEN %LET OverWrit = %UPCASE(&OverWrit);
  %IF ( &ChkExist NE ) %THEN %LET ChkExist = %UPCASE(&ChkExist);

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* Check arguments for legal contents and discrepancies, calling e_r_r_o_r_s    ;
*______________________________________________________________________________;

  %LET ErrCount = 0;
  %IF ( &SourceDs EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: SourceDs argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF ( %SYSFUNC(EXIST(&SourceDs)) EQ 0 ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Dataset &SourceDs does not exist,;
    %PUT                       macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE
  %DO; %* Check for non-existing variables within dataset;
   %IF ( %UPCASE(&ExclList) NE AND &ChkExist EQ Y ) %THEN
    %ChkVar (Data=&SourceDs, VarList=&ExclList, CalledBy=&MacName,
             Result = ErrCount); %* adding up non-existent variables ;
  %END;
  %IF ( &TargetDs EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: TargetDs argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF ( %SYSFUNC(EXIST(&TargetDs)) EQ 1 AND &OverWrit NE Y ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Dataset &TargetDs exists and OverWrit is not Y,;
    %PUT                       macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %IF ( &ErrCount GT 0 ) %THEN %GOTO Interupt;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Build InclList, exclude ExclList variables ;
%*____________________________________________;

%LET InclList = ; %* empty;
PROC CONTENTS DATA=&SourceDs %IF (&ChkExist EQ Y) %THEN (DROP=&ExclList);
               OUT=&Contents (KEEP=Name) NOPRINT;
RUN;

DATA _NULL_;
  SET &Contents;
  %IF (&ChkExist NE Y) %THEN
  %DO;
    %LET I = 1;
    %LET Id = %SCAN(&ExclList,&I,%STR( ));
    %DO %WHILE (&Id NE);
      IF UPCASE(Name) EQ UPCASE("&Id") THEN RETURN;
      %LET I = %EVAL( &I + 1 );
      %LET Id = %SCAN(&ExclList,&I,%STR( ));
    %END; %*=== While;
  %END;
  CALL EXECUTE ( '%LET Incllist = &Incllist ' || Name || ';' );
RUN;
%PUT &MacName: InclList=&InclList;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Automatic determination of attributes ;
%*_______________________________________;

PROC CONTENTS DATA=&SourceDs (KEEP=&InclList) OUT=&Contents NOPRINT;
RUN;

DATA _NULL_;
  SET &Contents (KEEP=Name Type);

 %LET I = 1;
 %LET Variable = %SCAN(&InclList,&I,%STR( ));
 %DO %WHILE (&Variable NE);

   %IF ( &I GT 1 ) %THEN ELSE;
    IF ( UPCASE(Name) = UPCASE("&Variable") ) THEN
    DO;
%* Make attributes as macro variables;
     %LOCAL MR&I.T; %* Define locals;
      CALL SYMPUT ("MR&I.T", TRIM(LEFT(PUT(Type,1.0))));
    END;

    %LET I = %EVAL( &I + 1 );
    %LET Variable = %SCAN(&InclList,&I,%STR( ));
  %END; %*=== While;

RUN;

PROC DATASETS NOLIST;
  DELETE &Contents;                        %* Delete scratch dataset &Contents;
RUN;

%IF ( &ErrCount GT 0 ) %THEN %GOTO Interupt;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Macro actually starts here ;
%*____________________________;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Check for a single non-missing value at maximum across multiple records ;
%*_________________________________________________________________________;

%LET DropList = ;
%LET I = 1;
%LET Variable = %SCAN(&InclList,&I,%STR( ));
%DO %WHILE (&Variable NE);

  PROC FREQ DATA=&SourceDs (KEEP=&Variable);
    TABLES &Variable / NOPRINT NOPERCENT OUT=&Scratch (KEEP=&Variable);
    RUN;

  DATA _NULL_;
    SET &Scratch NOBS=Nobs;
    %IF (&&MR&I.T EQ 1) %THEN /* numeric variable */
    %DO;
      IF (&Variable LE .z AND Nobs=1) THEN
    %END;
    %ELSE /* character variable */
    %DO;
      IF (&Variable = ' ' AND Nobs=1) THEN
    %END;
        CALL EXECUTE ( '%LET DropList = &DropList ' || "&Variable" || ';' );
    STOP;
  RUN;

  %LET I = %EVAL( &I + 1 );
  %LET Variable = %SCAN(&InclList,&I,%STR( ));
%END; %*=== While;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Remove redundant empty variables ;
%*__________________________________;

%PUT &MacName: DropList=&DropList;

DATA &TargetDs;
  SET &SourceDs %IF (&DropList NE) %THEN (DROP=&DropList);;
RUN;

%Aborting:

PROC DATASETS NOLIST;
  DELETE &Scratch;   %* Delete scratch datasets;
RUN;

%Interupt:

 %IF ( &ErrCount GT 0 ) %THEN
 %DO;
   DATA _NULL_;
     ABORT /*ABEND*/;
   RUN;
 %END;

%Finish:

%***********;
%MEND DROPMV; * --------------------end-of-macro-DROPMV------------------------;
************;

